from ordeq_common.io.bytes_buffer import BytesBuffer
from ordeq_common.io.dataclass import Dataclass
from ordeq_common.io.iterate import Iterate
from ordeq_common.io.literal import Literal
from ordeq_common.io.match import Match, MatchOnLoad, MatchOnSave
from ordeq_common.io.printer import Print
from ordeq_common.io.string_buffer import StringBuffer

__all__ = (
    "BytesBuffer",
    "Dataclass",
    "Iterate",
    "Literal",
    "Match",
    "MatchOnLoad",
    "MatchOnSave",
    "Print",
    "StringBuffer",
)



from ordeq_common.hooks import LoggerHook, SpyHook
from ordeq_common.io import (
    BytesBuffer,
    Dataclass,
    Iterate,
    Literal,
    Match,
    MatchOnLoad,
    MatchOnSave,
    Print,
    StringBuffer,
)

__all__ = (
    "BytesBuffer",
    "Dataclass",
    "Iterate",
    "Literal",
    "LoggerHook",
    "Match",
    "MatchOnLoad",
    "MatchOnSave",
    "Print",
    "SpyHook",
    "StringBuffer",
)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import DataclassInstance
from dataclasses import dataclass

from ordeq import IO, Input


@dataclass(frozen=True)
class Dataclass(Input["DataclassInstance"]):
    """IO that parses data as Python dataclass on load.

    Example:

    ```pycon
    >>> from ordeq_common import Dataclass
    >>> from ordeq_files import JSON
    >>> from pathlib import Path
    >>> ValidJSON = JSON(path=Path("to/valid.json"))
    >>> ValidJSON.load()  # doctest: +SKIP
    {"name": "banana", "colour": "yellow"}

    >>> @dataclass
    ... class Fruit:
    ...     name: str
    ...     colour: str
    >>> Dataclass(ValidJSON, Fruit).load()  # doctest: +SKIP
    Fruit(name="banana", colour="yellow")

    >>> InvalidJSON = JSON(path=Path("to/invalid.json"))
    >>> InvalidJSON.load()  # doctest: +SKIP
    {"name": "banana", "weight_gr": "100"}
    >>> Dataclass(InvalidJSON, Fruit).load()  # doctest: +SKIP
    TypeError: Fruit.__init__() got an unexpected keyword argument 'weight_gr'
    ```

    For nested models, or other more sophisticated parsing requirements
    consider using `ordeq-pydantic` instead.

    """

    io: IO[dict]
    dataclass: type[DataclassInstance]

    def load(self) -> DataclassInstance:
        data = self.io.load()
        # noinspection PyArgumentList
        return self.dataclass(**data)



from dataclasses import dataclass
from typing import Any

from ordeq import Output


@dataclass(frozen=True, eq=False)
class Print(Output[Any]):
    """Output that prints data on save. Mostly useful for debugging purposes.
    The difference between other utilities like `StringBuffer` and `Pass` is
    that `Print` shows the output of the node directly on the console.

    Example:

    ```pycon
    >>> from ordeq_common import Print, Literal
    >>> from ordeq import node, run
    >>> @node(
    ...     inputs=Literal("hello, world!"),
    ...     outputs=Print()
    ... )
    ... def print_message(message: str) -> str:
    ...     return message.capitalize()

    >>> run(print_message)
    Hello, world!

    >>> import sys
    >>> @node(
    ...     inputs=Literal("error message"),
    ...     outputs=Print().with_save_options(file=sys.stderr)
    ... )
    ... def log_error(message: str) -> str:
    ...     return f"Error: {message}"

    >>> run(log_error)  # prints to stderr

    ```
    """

    def save(self, data: Any, **save_options: Any) -> None:
        print(data, **save_options)



from dataclasses import dataclass
from typing import TypeVar

from ordeq import Input

T = TypeVar("T")


@dataclass(frozen=True, eq=False)
class Literal(Input[T]):
    """IO that returns a pre-defined value on load. Mostly useful for
    testing purposes.

    Example:

    ```pycon
    >>> from ordeq_common import Literal
    >>> value = Literal("someValue")
    >>> value.load()
    'someValue'
    >>> print(value)
    Literal('someValue')

    ```

    """

    value: T

    def load(self) -> T:
        return self.value

    def __repr__(self):
        return f"Literal({self.value!r})"



from collections.abc import Iterable
from dataclasses import dataclass
from typing import TypeVar

from ordeq import IO

T = TypeVar("T")


@dataclass(frozen=True)
class _Iterate(IO[Iterable[T]]):
    ios: tuple[IO[T], ...]

    def load(self) -> Iterable[T]:
        for io in self.ios:
            yield io.load()

    def save(self, data: Iterable[T]) -> None:
        for io, contents in zip(self.ios, data, strict=True):
            io.save(contents)


def Iterate(*ios: IO[T]) -> _Iterate[T]:
    """IO for loading and saving iteratively. This can be useful when
    processing multiple IOs using the same node, while only requiring
    to have one of them in memory at the same time.

    Examples:

    The load function returns a generator:

    ```pycon
    >>> from pathlib import Path
    >>> from ordeq_files import Text, JSON
    >>> from ordeq_common import Iterate
    >>> paths = [Path("hello.txt"), Path("world.txt")]
    >>> text_ios = Iterate(*[Text(path=path) for path in paths])
    >>> text_ios.load()  # doctest: +SKIP
    <generator object Iterate._load at 0x104946f60>
    ```

    The load function returns the contents of the files in this case:

    ```pycon
    >>> list(text_ios.load())  # doctest: +SKIP
    ['hello', 'world']
    ```

    By iterating over the contents, each file will be loaded and saved
    without the need to keep multiple files in memory at the same time:

    ```pycon
    >>> for idx, content in enumerate(text_ios.load()):   # doctest: +SKIP
    ...    JSON(
    ...        path=paths[idx].with_suffix(".json")
    ...    ).save({"content": content})

    ```

    We can achieve the same by passing a generator to the `Iterate.save`
    method:

    ```pycon
    >>> json_dataset = Iterate(
    ...     *[
    ...         JSON(path=path.with_suffix(".json"))
    ...         for path in paths
    ...     ]
    ... )
    >>> json_dataset.save(
    ...    ({"content": content} for content in text_ios.load())
    ... )   # doctest: +SKIP
    >>> from collections.abc import Iterable
    >>> def generate_json_contents(
    ...     contents: Iterable[str]
    ... ) -> Iterable[dict[str, str]]:
    ...     for content in contents:
    ...         yield {"content": content}
    >>> json_dataset.save(generate_json_contents(text_ios.load()))   # doctest: +SKIP

    ```

    Returns:
        _Iterate

    """  # noqa: E501 (line too long)

    return _Iterate(ios=tuple(ios))



from dataclasses import InitVar, dataclass, field
from io import StringIO

from ordeq import IO


@dataclass(frozen=True, eq=False)
class StringBuffer(IO[str]):
    """IO that uses an in-memory string buffer to load and save data. Useful
    for buffering data across nodes without writing to disk.

    Example:

    ```pycon
    >>> from ordeq_common import StringBuffer
    >>> buffer = StringBuffer()
    >>> buffer.load()
    ''

    ```

    The buffer is initially empty, unless provided with initial data:

    ```pycon
    >>> buffer = StringBuffer("Initial data")
    >>> buffer.load()
    'Initial data'

    ```

    Saving to the buffer appends data to the existing content:

    ```pycon
    >>> buffer.save("New data")
    >>> buffer.load()
    'Initial dataNew data'

    ```

    Example in a node:

    ```pycon
    >>> from ordeq_args import CommandLineArg
    >>> from ordeq_common import StringBuffer, Literal
    >>> from ordeq import node, run
    >>> result = StringBuffer()
    >>> @node(
    ...     inputs=[StringBuffer("Hello"), Literal("you")],
    ...     outputs=result
    ... )
    ... def greet(greeting: str, name: str) -> str:
    ...     return f"{greeting} to {name}!"
    >>> _ = run(greet)
    >>> result.load()
    'Hello to you!'

    ```

    """

    _buffer: StringIO = field(default_factory=StringIO, init=False)
    value: InitVar[str | None] = None

    def __post_init__(self, value: str | None):
        if value is not None:
            self._buffer.write(value)

    def load(self) -> str:
        return self._buffer.getvalue()

    def save(self, data: str) -> None:
        self._buffer.write(data)

    def persist(self, _) -> None:
        return super().persist(self._buffer.getvalue())



from dataclasses import dataclass, replace
from typing import Generic, TypeVar, overload

from ordeq import IO, Input, Output

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

T = TypeVar("T")
Tkey = TypeVar("Tkey")
Tval = TypeVar("Tval")


@dataclass(frozen=True, kw_only=True)
class MatchOnLoad(Input[Tval], Generic[Tkey, Tval]):
    io: Input[Tkey]
    cases: tuple[tuple[Tkey, Input[Tval] | IO[Tval]], ...] = ()
    default: Input[Tval] | IO[Tval] | None = None

    def _match(self, value: Tkey) -> Input[Tval] | IO[Tval]:
        for case, io in self.cases:
            if value == case:
                return io
        if self.default is not None:
            return self.default
        raise TypeError(f"Unsupported case '{value}'")

    def load(self) -> Tval:
        return self._match(self.io.load()).load()

    def Case(self, key: Tkey, val: Input[Tval] | IO[Tval]) -> Self:
        return replace(self, cases=(*self.cases, (key, val)))

    def Default(self, val: Input[Tval] | IO[Tval]) -> Self:
        return replace(self, default=val)


@dataclass(frozen=True, kw_only=True)
class MatchOnSave(Output[tuple[Tkey, Tval]], Generic[Tkey, Tval]):
    cases: tuple[tuple[Tkey, Output[Tval] | IO[Tval]], ...] = ()
    default: Output[Tval] | IO[Tval] | None = None

    def _match(self, value: Tkey) -> Output[Tval] | IO[Tval]:
        for case, io in self.cases:
            if value == case:
                return io
        if self.default is not None:
            return self.default
        raise TypeError(f"Unsupported case '{value}'")

    def save(self, data: tuple[Tkey, Tval]) -> None:
        key, value = data
        self._match(key).save(value)

    def Case(self, key: Tkey, val: Output[Tval] | IO[Tval]) -> Self:
        return replace(self, cases=(*self.cases, (key, val)))

    def Default(self, val: Output[Tval] | IO[Tval]) -> Self:
        return replace(self, default=val)


@overload
def Match(io: Input[Tkey] | IO[Tkey]) -> MatchOnLoad[Tval, Tkey]: ...


@overload
def Match() -> MatchOnSave[Tval, Tkey]: ...


def Match(
    io: Input[Tkey] | IO[Tkey] | None = None,
) -> MatchOnLoad | MatchOnSave:
    """Utility IO that allows dynamic switching between IO, like the match-case
    statement in Python.

    Example:

    ```pycon
    >>> from ordeq_common import Literal, Match
    >>> from ordeq_args import EnvironmentVariable
    >>> import os
    >>> Country = (
    ...     Match(EnvironmentVariable("COUNTRY"))
    ...     .Case("NL", Literal("Netherlands"))
    ...     .Case("BE", Literal("Belgium"))
    ...     .Default(Literal("Unknown"))
    ... )
    >>> os.environ["COUNTRY"] = "NL"
    >>> Country.load()
    'Netherlands'

    ```

    If a default is provided, it will be used when no cases match:

    ```pycon
    >>> os.environ["COUNTRY"] = "DE"
    >>> Country.load()
    'Unknown'

    ```

    Otherwise, it raises an error when none of the provided cases are matched:

    ```pycon
    >>> Match(EnvironmentVariable("COUNTRY")).load()  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ordeq.IOException: Failed to load
    Unsupported case 'DE'
    ```

    Match on save works as follows:

    ```pycon
    >>> SmallOrLarge = (
    ...     Match()
    ...     .Case("S", EnvironmentVariable("SMALL"))
    ...     .Case("L", EnvironmentVariable("LARGE"))
    ...     .Default(EnvironmentVariable("UNKNOWN"))
    ... )
    >>> SmallOrLarge.save(("S", "Andorra"))
    >>> SmallOrLarge.save(("L", "Russia"))
    >>> SmallOrLarge.save(("XXL", "Mars"))
    >>> os.environ["SMALL"]
    'Andorra'
    >>> os.environ.get("LARGE")
    'Russia'
    >>> os.environ.get("UNKNOWN")
    'Mars'

    ```

    Example in a node:

    ```pycon
    >>> from ordeq import node
    >>> from ordeq_files import JSON
    >>> from ordeq_args import CommandLineArg
    >>> from pathlib import Path
    >>> TestOrTrain = (
    ...     Match(CommandLineArg("--split"))
    ...     .Case("test", JSON(path=Path("to/test.json")))
    ...     .Case("train", JSON(path=Path("to/train.json")))
    ... )
    >>> @node(
    ...     inputs=TestOrTrain,
    ... )
    ... def evaluate(data: dict) -> dict:
    ...     ...

    ```

    Returns:
        MatchOnLoad or MatchOnSave
    """  # noqa: E501 (line too long)

    if io is None:
        return MatchOnSave()
    return MatchOnLoad(io=io)



from dataclasses import InitVar, dataclass, field
from io import BytesIO

from ordeq import IO


@dataclass(frozen=True, eq=False)
class BytesBuffer(IO[bytes]):
    """IO that uses an in-memory bytes buffer to load and save data. Useful
    for buffering data across nodes without writing to disk.

    Example:

    ```pycon
    >>> from ordeq_common import BytesBuffer
    >>> buffer = BytesBuffer()
    >>> buffer.load()
    b''

    ```

    The buffer is initially empty, unless provided with initial data:

    ```pycon
    >>> buffer = BytesBuffer(b"Initial data")
    >>> buffer.load()
    b'Initial data'

    ```

    Saving to the buffer appends data to the existing content:

    ```pycon
    >>> buffer.save(b"New data")
    >>> buffer.load()
    b'Initial dataNew data'

    ```

    Example in a node:

    ```pycon
    >>> from ordeq_args import CommandLineArg
    >>> from ordeq_common import BytesBuffer, Literal
    >>> from ordeq import node, run
    >>> result = BytesBuffer()
    >>> @node(
    ...     inputs=[BytesBuffer(b"Hello"), Literal(b"you")], outputs=result
    ... )
    ... def greet(greeting: bytes, name: bytes) -> bytes:
    ...     return greeting + b" to " + name + b"!"
    >>> _ = run(greet)
    >>> result.load()
    b'Hello to you!'

    ```

    """

    _buffer: BytesIO = field(default_factory=BytesIO, init=False)
    value: InitVar[bytes | None] = None

    def __post_init__(self, value: bytes | None):
        if value is not None:
            self._buffer.write(value)

    def load(self) -> bytes:
        return self._buffer.getvalue()

    def save(self, data: bytes) -> None:
        self._buffer.write(data)

    def persist(self, _) -> None:
        return super().persist(self._buffer.getvalue())



import logging
from logging import Logger, getLogger

from ordeq import InputHook, NodeHook, OutputHook

_DEFAULT_LOGGER = getLogger("LoggerHook")
_DEFAULT_LOGGER.setLevel(logging.INFO)


class LoggerHook(InputHook, OutputHook, NodeHook):
    """Hook that prints the calls to the methods.
    Typically only used for test purposes."""

    def __init__(
        self, *, logger: Logger = _DEFAULT_LOGGER, level: int | None = None
    ) -> None:
        self._logger = logger
        self._level = logger.level
        if level is not None:
            self._level = level

    def _register_call(self, item, *args) -> None:
        self._logger.log(self._level, "Called '%s' with args: %s", item, args)

    def before_node_run(self, *args) -> None:
        return self._register_call("before_node_run", *args)

    def before_input_load(self, *args) -> None:
        return self._register_call("before_input_load", *args)

    def after_input_load(self, *args) -> None:
        return self._register_call("after_input_load", *args)

    def on_node_call_error(self, *args) -> None:
        return self._register_call("on_node_call_error", *args)

    def before_output_save(self, *args) -> None:
        return self._register_call("before_output_save", *args)

    def after_output_save(self, *args) -> None:
        return self._register_call("after_output_save", *args)

    def after_node_run(self, *args) -> None:
        return self._register_call("after_node_run", *args)



from ordeq import InputHook, NodeHook, OutputHook


class SpyHook(InputHook, OutputHook, NodeHook):
    """Hook that stores the arguments it is called with in a list. Typically
    only used for test purposes."""

    def __init__(self):
        self.called_with: list[tuple] = []

    def _register_call(self, item, *args) -> None:
        self.called_with.append((item, *args))

    def before_node_run(self, *args) -> None:
        return self._register_call("before_node_run", *args)

    def before_input_load(self, *args) -> None:
        return self._register_call("before_input_load", *args)

    def after_input_load(self, *args) -> None:
        return self._register_call("after_input_load", *args)

    def on_node_call_error(self, *args) -> None:
        return self._register_call("on_node_call_error", *args)

    def before_output_save(self, *args) -> None:
        return self._register_call("before_output_save", *args)

    def after_output_save(self, *args) -> None:
        return self._register_call("after_output_save", *args)

    def after_node_run(self, *args) -> None:
        return self._register_call("after_node_run", *args)
